/*********************************************************************/
/* file: llist.c - linked-list datastructure                         */
/*                             HUT TINTIN                            */
/*                coded by Ossi Tulonen from TinTin III              */
/*                TinTin III coded by peter unold 1992               */
/*********************************************************************/
/*
 * TinTin III code Copyright (c) 1992 Peter Unold
 * All changed code from TinTin III Copyright (c) 1992 1993 1994 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 
#include <stdio.h>
#include <string.h>
#ifndef NSTDLIBS
#include <stdlib.h>
#endif

#include "structs.h"
#include "color.h"
#include "llist.h"
#include "parse.h"
#include "screen.h"
#include "utility.h"
#include "utils.h"

extern struct master_table_data master_table;
extern unsigned char text_start, text_end;

extern const uchar *RESETCOLOR;

/*****************************************************************/
/* create a node containing the ltext, rtext fields and stuff it */
/* into the list - in lexicographical order                      */
/*****************************************************************/

listnode::listnode(listnode **olisthead,
		   const uchar *ltext, const uchar *rtext,
		   const int otype, const int opri) : type(otype), pri(opri)
{
  listnode *nptr, *nptrlast;
  
  listhead = olisthead;
  left  = (uchar *)strdup((const char *)ltext);
  right = (uchar *)strdup((const char *)rtext);
  
  if(*listhead) {
    if(pri < (*listhead)->pri ||
       (pri == (*listhead)->pri &&
	strcmp((const char *)ltext, (const char *)(*listhead)->left) <= 0)) {
      next  = *listhead;
      *listhead = this;
      return;
    }
    for(nptr = *listhead, nptrlast = nptr;
	nptr && nptr->pri < pri;
	nptrlast = nptr, nptr = nptr->next)
      ;
    
    for(; nptr && nptr->pri == pri; nptrlast = nptr, nptr = nptr->next) {
      if(strcmp((const char *)ltext, (const char *)nptr->left) <= 0)
        break;
    }
    nptrlast->next = this;
    next  = nptr;
  } else {
    *listhead = this;
    next  = 0;
  }
}

/*****************************/
/* delete a node from a list */
/*****************************/
listnode::~listnode()
{
  listnode *lastnode, *tmp;

  if(*listhead == this) {
    *listhead = next;
    if(left)
      free(left);
    if(right)
      free(right);
  } else {
    for(lastnode = *listhead, tmp = (*listhead)->next;
	tmp;
	lastnode = tmp, tmp = tmp->next) {
      if(tmp == this) {
	lastnode->next = next;
	if(left)
	  free(left);
	if(right)
	  free(right);
	return;
      }
    }
  }
}

/************************************/
/* show contens of a node on screen */
/************************************/
void listnode::show(session *ses) const
{
  unsigned char buf[BUFFER_SIZE], tmp[BUFFER_SIZE];
  
  switch(type) {
  case TYPE_COLOR:
    color_to_text(right, tmp, TRUE);
    sprintf((char *)buf, "%c%s%s%s%c=%c%s%c\n\r",
	    text_start, right, left, RESETCOLOR, text_end,
	    text_start, tmp, text_end);
    break;
  case TYPE_ACTION:
    sprintf((char *)buf, "A(%2d) %c%s%c=%c%s%c\n\r",
	    pri,
	    text_start, left, text_end,
	    text_start, right, text_end);
    break;
  case TYPE_SUBS:
    sprintf((char *)buf, "S(%2d) %c%s%c=%c%s%c\n\r",
	    pri,
	    text_start, left, text_end,
	    text_start, right, text_end);
    break;
  case TYPE_GAGS:
    sprintf((char *)buf, "G(%2d) %c%s%c\n\r",
	    pri, text_start, left, text_end);
    break;
  case TYPE_ALIAS:
  case TYPE_VAR:
    sprintf((char *)buf, "%c%s%c=%c%s%c\n\r",
	    text_start, left, text_end,
	    text_start, right, text_end);
    break;
  default:
    fprintf(stderr, "#BUG: Unknown listnode type.\n");
    break;
  }
  
  ses->print(buf);
}

/************************************************/
/* kill list - run throught list and free nodes */
/************************************************/ 
void kill_list(listnode **nptr)
{
  listnode *tmp, *nexttodel;

  for(tmp = *nptr; tmp; tmp = nexttodel) {
    nexttodel = tmp->next;
    delete tmp;
  }
  *nptr = 0;
}

/***********************************************/
/* make a copy of a list - return: ptr to copy */
/***********************************************/
void copy_list(listnode *sourcelist, listnode **resultlist)
{
  listnode *tmp;
  
  *resultlist = 0;
  while(sourcelist) {
    tmp = new listnode(resultlist, sourcelist->left, sourcelist->right,
		       sourcelist->type, sourcelist->pri);
    sourcelist = sourcelist->next;
  }
}

/********************************************************/
/* search for a node containing the ltext in left-field */
/* return: ptr to node on succes / NULL on failure      */
/********************************************************/
listnode *searchnode_list(listnode *listhead,
			  unsigned char *cptr)
{
  int i;
  
  while(listhead) {
    if((i = strcmp((const char *)listhead->left, (const char *)cptr)) == 0)
      return listhead;
    
    listhead = listhead->next;
  }
  return NULL;
}

/********************************************************/
/* search for a node containing the ltext in left-field */
/* return: ptr to node on succes / NULL on failure      */
/********************************************************/
listnode *searchnode_list2(listnode *listhead,
			   unsigned char *cptr)
{
  while(listhead) {
    if(mystrcmp(listhead->left, cptr))
      return listhead;
    
    listhead = listhead->next;
  }
  return NULL;
}

/************************************/
/* list contens of a list on screen */
/************************************/
void show_list(listnode *listhead, session *ses)
{
  while(listhead) {
    listhead->show(ses);
    listhead = listhead->next;
  }
}
