/*********************************************************************/
/* file: session.c.c - funtions related to sessions                  */
/*                             HUT TINTIN                            */
/*                coded by Ossi Tulonen from TinTin III              */
/*                TinTin III coded by peter unold 1992               */
/*********************************************************************/
/*
 * TinTin III code Copyright (c) 1992 Peter Unold
 * All changed code from TinTin III Copyright (c) 1992 1993 1994 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#ifdef OSF1v3
typedef int clockid_t;
#endif
#include <time.h>
#ifndef NSTDLIBS
#include <stdlib.h>
#include <unistd.h>
#endif
#include <sys/types.h>

#include "structs.h"
#include "files.h"
#include "llist.h"
#include "main.h"
#include "net.h"
#include "parse.h"
#include "screen.h"
#include "session.h"
#ifdef TERM
#  include "term.h"
#endif
#include "utility.h"
#include "utils.h"

#ifdef TERM
#  define CLOSE(sock)                    term_close(sock)
#else
#  define CLOSE(sock)                    close(sock)
#endif

session *session::sessions = 0;

extern int quietmode;
extern struct master_table_data master_table;

extern const uchar *CEL;

// ***********************
// class session functions
// ***********************

session::session()
{
  int i;
  
  name         = (uchar *)strdup("master");
  address      = (uchar *)strdup("your local machine");
  autolog      = 0;
  for(i = 0; i < WORD_LIST_SIZE; i++)
    word_list[i] = 0;

  tickstatus   = FALSE;
  tick_size    = 75;
  sec_to_tick  = 75;
  start_time   = time(0);
  snoopstatus  = FALSE;
  echo_flags   = DEFAULT_ECHO;
  message_flags= DEFAULT_MESS;
  line_mode    = TRUE;
  ignore_flags = 0;
  logfile      = 0;
  aliases      = 0;
  actions      = 0;
  colors       = 0;
  variables    = 0;
  path_mark    = 0;
  path_now     = 0;
  path_length  = 0;
  for(i = 0; i < NTELOPTS; i++) {
    will_status[i] = 0;
    do_status[i] = 0;
  }
  for(i = 0; i < MAX_SES_HIST; i++)
    history[i] = 0;
  session_hist_end = 0;
  session_hist_num = 0;
  session_hist_mark_pos = 0;
  session_hist_mark_col = 1;
  current_line[0] = '\0';
  current_pos     = 0;

  promptbuf[0] = '>';
  promptbuf[1] = ' ';
  promptbuf[2] = '\0';
  prompt_pos   = 2;
  waitcomms    = 0;
  net          = -1;
  term_echoing = TRUE;
  input        = window::get_active_win();
  output       = window::get_active_win();
  input->set_active_ses(this);

  next         = 0;
}

session::session(const unsigned char *nam, const unsigned char *add,
		 const int sock)
{
  int i;

  name         = (uchar *)strdup((const char *)nam);
  address      = (uchar *)strdup((const char *)add);
  autolog      = 0;
  for(i = 0; i < WORD_LIST_SIZE; i++) {
    if(master_table.main_session->word_list[i])
      word_list[i] = (uchar *)strdup((const char *)master_table.main_session->word_list[i]);
    else
      word_list[i] = 0;
  }
  tickstatus   = FALSE;
  tick_size    = 75;
  sec_to_tick  = 75;
  start_time   = time(0);
  snoopstatus  = FALSE;
  echo_flags   = master_table.main_session->echo_flags;
  message_flags= master_table.main_session->message_flags;
  line_mode    = TRUE;
  ignore_flags = master_table.main_session->ignore_flags;
  logfile      = 0;
  copy_list(master_table.main_session->aliases,   &aliases);
  copy_list(master_table.main_session->actions,   &actions);
  copy_list(master_table.main_session->colors,    &colors);
  copy_list(master_table.main_session->variables, &variables);
  path_mark    = 0;
  path_now     = 0;
  path_length  = 0;
  for(i = 0; i < NTELOPTS; i++) {
    will_status[i] = 0;
    do_status[i] = 0;
  }
  for(i = 0; i < MAX_SES_HIST; i++)
    history[i] = 0;
  session_hist_end = 0;
  session_hist_num = 0;
  session_hist_mark_pos = 0;
  session_hist_mark_col = 1;
  current_line[0] = '\0';
  current_pos     = 0;
  *promptbuf   = '\0';
  prompt_pos   = 0;
  waitcomms    = 0;
  net          = sock;
  term_echoing = TRUE;
  input        = window::get_active_win();
  output       = window::get_active_win();
  
  next              = session::sessions;
  session::sessions = this;
  input->set_active_ses(this);
  master_table.main_session->next = this;
}

session::~session()
{
  unsigned char buf[BUFFER_SIZE];
  int i;
  session *sesptr;
  struct waitcommdata *tmp, *tmpnext;
  session *newactiv, *ses;
  window *win;

  if(autolog) {
    comm_logseshist(autolog, this);

    free(autolog);
    autolog = 0;
  }
  if(this == session::sessions)
    session::sessions = next;
  else {
    if((sesptr = session::sessions)) {
      while(sesptr->next && sesptr->next != this)
	sesptr = sesptr->next;
      
      if(sesptr->next)
	sesptr->next = next;
    }
  }
  
  for(win = window::get_windows(); win; win = win->get_next()) {
    if(win->get_hist_on() == this) {
      win->exit_win_history();
    }
  }
  
  if(net >= 0)
    if(CLOSE(net) < 0)
      perror("#close in cleanup_session");
  
  for(i = 0; i < WORD_LIST_SIZE; i++)
    if(word_list[i]) {
      free(word_list[i]);
      word_list[i] = 0;
    }
  
  for(i = 0; i < MAX_SES_HIST; i++)
    if(history[i]) {
      free(history[i]);
      history[i] = 0;
    }

  kill_list(&aliases);
  kill_list(&actions);
  kill_list(&colors);
  kill_list(&variables);

  for(tmp = waitcomms; tmp; tmp = tmpnext) {
    tmpnext = tmp->next;
    free(tmp->comm);
    free(tmp);
  }

  sprintf((char *)buf, "#SESSION '%s' DIED.\n\r", name);
  print(buf);

  if(name)
    free(name);
  if(address)
    free(address);
  if(logfile)
    fclose(logfile);
  
  if(this == master_table.main_session) {
    master_table.main_session = 0;
    exit_tintin(0);
  } else {
    if(master_table.main_session)
      master_table.main_session->next = session::sessions;
  }
  
  if(this == input->get_active_ses()) {
    for(newactiv = session::get_sessions();
	newactiv;
	newactiv = newactiv->get_next()) {
      if(newactiv->get_input() == input) {
	break;
      }
    }
    if(!newactiv) {
      if(master_table.main_session) {
	newactiv = master_table.main_session;
	if(newactiv->get_input()->get_active_ses() == newactiv) {
	  for(ses = session::get_sessions(); ses; ses = ses->get_next())
	    if(ses->get_input() == newactiv->get_input())
	      break;
	  
	  newactiv->get_input()->set_active_ses(ses);
	}
	input->set_active_ses(newactiv);
	newactiv->set_input(input);
	if(!window::get_active_win()->get_active_ses()) {
	  window::set_active_win(input);
	}
      }
    } else {
      input->set_active_ses(newactiv);
    }
    if(newactiv) {
      sprintf((char *)buf, "#SESSION '%s' ACTIVATED.\n\r", newactiv->get_name());
      newactiv->print(buf);
      newactiv->get_input()->set_ptexrem();
    }
  }
  input->set_ptexrem();
}

void session::print(const unsigned char *text)
{
  unsigned char buf[BUFFER_SIZE];
  int pos;
  window *win;
  
  if(quietmode) {
    return;
  }
  
  if(*text) {
    insert_ses_scrhist(text, FALSE);
    
    if(master_table.output_window) {
      if(master_table.output_window == NULL_WIN)
	return;
      
      win = master_table.output_window;
    } else
      win = output;
    
    if(!win->get_hist_on()) {
      if((this == win->get_active_ses()) ||
	 (!win->get_active_ses() &&
	  window::get_active_win()->get_active_ses() == this)) {
	win->remove_user_text();
	win->print(text);
      } else if(snoopstatus || !win->get_active_ses()) {
	win->remove_user_text();
	pos = 0;
	pos = str_cat(buf, pos, sizeof(buf), (const uchar *)"\r");
	pos = str_cat(buf, pos, sizeof(buf), CEL);
	win->print(buf);
	pos = 0;
	pos = str_cat(buf, pos, sizeof(buf), name);
	pos = str_cat(buf, pos, sizeof(buf), (const uchar *)"% ");
	win->print(text, buf);
	win->set_ptexrem();
	win->print_prompt(TRUE, FALSE);
      }
    }
  }
}

void session::print_prompt(const int all)
{
  if(this == output->get_active_ses()) {
    output->print_prompt(all);
  }
}

const uchar *session::get_prompt() const
{
  static uchar prompt_store[MAX_PROMPT_SIZE + 2];
  int pos;
  
  if(master_table.inputbuf2_pos) {
    pos = str_cat(prompt_store, 0, sizeof(prompt_store), promptbuf);
    pos = str_cat(prompt_store, pos, sizeof(prompt_store), (const uchar *)"+ ");
    return(prompt_store);
  }
  return(promptbuf);
}

const int session::get_prompt_len() const
{
  if(master_table.inputbuf2_pos) {
    return(prompt_pos + 2);
  }
  return prompt_pos;
}

void session::set_prompt(const unsigned char *text)
{
  str_cat(promptbuf, 0, MAX_PROMPT_SIZE, text);
  prompt_pos = str_len(promptbuf);
}

const unsigned char *session::search_word(const unsigned char *word) const
{
  int pos;
  
  for(pos = 0; pos < WORD_LIST_SIZE && word_list[pos]; pos++)
    if(strcmp((const char *)word, (const char *)word_list[pos]) == 0)
      break;
  
  if(pos < WORD_LIST_SIZE && word_list[pos]) {
    return(word_list[pos]);
  }
  return(0);
}

void session::set_autolog(const unsigned char *name)
{
  if(autolog) {
    free(autolog);
    autolog = 0;
  }
  if(name) {
    autolog = (uchar *)strdup((const char *)name);
  }
}

/************************/
/* the #session command */
/************************/
session *session_command(const unsigned char *arg, session *ses)
{
  unsigned char left[BUFFER_SIZE], right[BUFFER_SIZE];
  session *sesptr;
  window *win;

  arg = get_arg_stop_spaces(arg, left);
  arg = get_arg_with_spaces(arg, right);

  if(!*left) {
    ses->print((const uchar *)"#THESE SESSIONS HAS BEEN DEFINED:\n\r");
    for(sesptr = master_table.main_session;
	sesptr;
	sesptr = sesptr->get_next())
      show_session(sesptr, ses);
  } else if(*left && !*right) {
    for(sesptr = master_table.main_session;
	sesptr;
	sesptr = sesptr->get_next())
      if(!strcmp((const char *)sesptr->get_name(), (const char *)left)) {
        show_session(sesptr, ses);
        break;
      }
    if(sesptr == NULL) {
      ses->print((const uchar *)"#THAT SESSION IS NOT DEFINED.\n\r");
    }
  } else {
    for(sesptr = master_table.main_session;
	sesptr;
	sesptr = sesptr->get_next())
      if(strcmp((const char *)sesptr->get_name(), (const char *)left) == 0) {
        ses->print((const uchar *)"#THERE'S A SESSION WITH THAT NAME ALREADY.\n\r");
        return 0;
      }
    win = find_window(left);
    if(win) {
      ses->print((const uchar *)"#THERE'S A WINDOW WITH THAT NAME ALREADY.\n\r");
      return 0;
    }
    
    ses = new_session(left, right, ses);
  }
  
  return ses;
}

/******************/
/* show a session */
/******************/
void show_session(session *show, session *ses) 
{
  unsigned char buf[BUFFER_SIZE];
  unsigned char len;

  sprintf((char *)buf, "%-10s (%2d)", show->get_name(), show->get_socket());
  len = str_cat(buf, strlen((const char *)buf), sizeof(buf), show->get_addr());

  if(show == show->get_input()->get_active_ses())
    len = str_cat(buf, len, sizeof(buf), (const uchar *)" (active)");
  if(show->snoopstatus)
    len = str_cat(buf, len, sizeof(buf), (const uchar *)" (snooped)");
  if(show->logfile)
    len = str_cat(buf, len, sizeof(buf), (const uchar *)" (logging)");
  if(show->message_flags) {
    len = str_cat(buf, len, sizeof(buf), (const uchar *)" (mess:");
    if(IS_SET(show->message_flags, MESSAGE_ACTION))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" AC");
    if(IS_SET(show->message_flags, MESSAGE_ALIAS))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" AL");
    if(IS_SET(show->message_flags, MESSAGE_SUBS))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" S");
    if(IS_SET(show->message_flags, MESSAGE_WAIT))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" W");
    if(IS_SET(show->message_flags, MESSAGE_VAR))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" V");
    len = str_cat(buf, len, sizeof(buf), (const uchar *)")");
  }
  if(show->echo_flags) {
    len = str_cat(buf, len, sizeof(buf), (const uchar *)" (echoed:");
    if(IS_SET(show->echo_flags, ECHO_ACTION))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" AC");
    if(IS_SET(show->echo_flags, ECHO_ALIAS))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" AL");
    if(IS_SET(show->echo_flags, ECHO_IF))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" I");
    if(IS_SET(show->echo_flags, ECHO_CMP))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" CM");
    if(IS_SET(show->echo_flags, ECHO_WAIT))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" W");
    len = str_cat(buf, len, sizeof(buf), (const uchar *)")");
  }
  if(show->ignore_flags) {
    len = str_cat(buf, len, sizeof(buf), (const uchar *)" (ignore:");
    if(IS_SET(show->ignore_flags, IGNORE_ACTION))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" AC");
    if(IS_SET(show->ignore_flags, IGNORE_SUBS))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" SU");
    if(IS_SET(show->ignore_flags, IGNORE_ALIAS))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" AL");
    if(IS_SET(show->ignore_flags, IGNORE_VAR))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" VA");
    if(IS_SET(show->ignore_flags, IGNORE_CALC))
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" CA");

    len = str_cat(buf, len, sizeof(buf), (const uchar *)")");
  }
  if(show->get_autolog()) {
    len = str_cat(buf, len, sizeof(buf), (const uchar *)"\n\r     (Autolog: ");
    len = str_cat(buf, len, sizeof(buf), show->get_autolog());
    len = str_cat(buf, len, sizeof(buf), (const uchar *)")");
  }
  len = str_cat(buf, len, sizeof(buf), (const uchar *)"\n\r");
  ses->print(buf);
}

/**********************/
/* open a new session */
/**********************/
session *new_session(unsigned char *name, unsigned char *address,
			    session *ses)
{
  int sock;
  unsigned char *host, *port, buf[1024], *defport = (uchar *)"23";
  session *newsession;

  while(*address == ' ')
    address++;

  for(host = address; *address != '\0' && *address != ' '; address++)
    ;
  if(*address) {
    *address = '\0';
    address++;
  }
  
  while(*address == ' ')
    address++;
  
  for(port = address; *address != '\0' && *address != ' '; address++)
    ;
  *address = '\0';

  if(!*host) {
    ses->print((const uchar *)"#HEY! SPECIFY AN ADDRESS WILL YOU?\n\r");
    return(0);
  }
  
  if(!*port) {
    port = defport;
    ses->print((const uchar *)"#Using default port number.\n\r");
  }
  
  if(!(sock = connect_mud(host, port, ses)))
    return(0);
  
  sprintf((char *)buf, "%s %s", host, port);
  newsession = new session(name, buf, sock);
  return(newsession);
}
